import base64
import logging
import os
import re
import sys
import time
import urllib.error
import urllib.parse
import urllib.request
from configparser import RawConfigParser

from cb.db.core_models import User
from cb.utils.config import Config
from cb.utils.db import DbSessionContext
from cb.utils import redis

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

mypath = os.path.dirname(os.path.abspath(__file__))
sys.path.append(mypath)

try:
    from duo_client import auth
except ImportError:
    log.critical("Could not import Duo client libraries")

cb_config = Config()
cb_config.load('/etc/cb/cb.conf')

# API calls to trap:
# POST /api/v*/cblr/session/*/command (cblr command endpoint)
# POST /api/auth (only for users that are 2fa enabled, though...)
# PUT /api/v*/sensor/* (network isolation)
# * /api/logout (clear our authentication tokens for that user)


class AuthenticationError(Exception):
    pass


class AuthenticationRequest(object):
    def __init__(self, user_object, typestring, short_type, path, method, ipaddr, duo_key, info=None):
        self.user_object = user_object
        self.typestring = typestring
        self.short_type = short_type
        self.path = path
        self.method = method
        self.ipaddr = ipaddr
        self.info = info
        self.duo_key = duo_key


class AuthenticationState(object):

    _DUO_SECRETS_DIR = '/etc/cb/plugins/duo'

    def __init__(self):
        self.r = redis.Redis('127.0.0.1')
        self.config = RawConfigParser()
        self.config.read(os.path.join(self._DUO_SECRETS_DIR, 'secrets.ini'))

        self.duo = auth.Auth(self.config.get("duo", "ikey"), self.config.get("duo", "skey"),
                             self.config.get("duo", "host"))

        self.duo_attribute = "email"
        if self.config.has_option("config", "duo_mapping"):
            self.duo_attribute = self.config.get("config", "duo_mapping")

        if self.duo_attribute not in ["email", "username"]:
            log.error("Duo attribute must be one of 'email' or 'username', defaulting to 'email'")
            self.duo_attribute = "email"

        self.session_lifetime = 60
        if self.config.has_option("config", "session_lifetime"):
            try:
                self.session_lifetime = int(self.config.get("config", "session_lifetime"))
            except ValueError:
                log.error("Session lifetime not an integer, defaulting to %d seconds" % self.session_lifetime)

        self.use_ipaddr_session_key = False
        if self.config.has_option("config", "use_ipaddr_session_key"):
            t = self.config.get("config", "use_ipaddr_session_key")
            if t.lower() == "true" or t == "1":
                self.use_ipaddr_session_key = True

        if self.config.has_option("proxy", "hostname") and self.config.has_option("proxy", "port"):
            proxy_header = None

            if self.config.has_option("proxy", "username") and self.config.has_option("proxy", "password"):
                proxy_user = self.config.get("proxy", "username")
                proxy_password = self.config.get("proxy", "password")

                token = base64.encodebytes('%s:%s'.encode('utf-8') % (proxy_user.encode('utf-8'),
                                                                      proxy_password.encode('utf-8'))).strip()
                proxy_header = {'Proxy-Authorization': 'Basic %s' % token}

            proxy_hostname = self.config.get("proxy", "hostname")
            proxy_port = int(self.config.get("proxy", "port"))

            self.duo.set_proxy(proxy_hostname, proxy_port, headers=proxy_header)

    def _get_keyname_for_user(self, request):
        if self.use_ipaddr_session_key:
            return "devr:2fa:%s:%s:%s" % (request.user_object.username, request.short_type, request.ipaddr)
        else:
            return "devr:2fa:%s:%s" % (request.user_object.username, request.short_type)

    def perform_2fa(self, user, path, method, ipaddr, request_data):
        # perform Duo
        # Allow Duo token users to perform 2fa and network isolation
        # Allow non-Duo token users to log in to the Web UI

        if is_auth_api_route(path, method):
            typestring = "Cb Web UI"
            short_type = "web"
            info = {"Requested Action": "Global Admin Web UI"}
        elif is_liveresponse_api_route(path, method, request_data):
            typestring = "Cb Live Response"
            short_type = "liveresponse"
            info = {"Requested Action": "Live Response Command"}
        else:
            typestring = "Other Cb API"
            short_type = "other"
            info = {"Requested Action": "Other Cb API"}

        if self.duo_attribute == "email":
            duo_key = user.email
        elif self.duo_attribute == "username":
            duo_key = user.username
        else:
            raise AuthenticationError("Invalid duo attribute: either 'email' or 'username' required. Failing login.")

        request = AuthenticationRequest(
            user_object=user,
            typestring=typestring,
            short_type=short_type,
            path=path,
            method=method,
            ipaddr=ipaddr,
            duo_key=duo_key,
            info=info
        )

        if not self.user_needs_2fa(request):
            return True

        self.required_2fa(request)
        return True

    def required_2fa(self, request):
        try:

            auth_result = self.duo.auth("auto", username=request.duo_key, type=request.typestring, device="auto",
                                        pushinfo=urllib.parse.urlencode(request.info),
                                        ipaddr=request.ipaddr)
        except RuntimeError as e_:
            raise AuthenticationError("Login request for %s from IP %s to path %s failed: %s" %
                                      (request.user_object.username,
                                       request.ipaddr,
                                       request.path,
                                       str(e_)))

        if auth_result.get("result", "deny") != "allow":
            raise AuthenticationError("Login request for %s from IP %s to path %s failed: %s" %
                                      (request.user_object.username,
                                       request.ipaddr,
                                       request.path,
                                       auth_result.get("status_msg", "Unknown status")))

        self.user_2fa_succeeded(request)
        return True

    def user_set_2fa_lifetime(self, request):
        keyname = self._get_keyname_for_user(request)
        self.r.set(keyname, int(time.time()))
        self.r.expire(keyname, self.session_lifetime)

    def user_needs_2fa(self, request):
        keyname = self._get_keyname_for_user(request)
        if self.r.exists(keyname):
            # renew expiration date
            self.user_set_2fa_lifetime(request)
            return False

        return True

    def user_2fa_succeeded(self, request):
        log.info("Duo 2-factor authentication succeeded for %s from IP %s to perform action %s" %
                 (request.user_object.username,
                  request.ipaddr,
                  request.short_type))
        self.user_set_2fa_lifetime(request)

    def clear_2fa(self, user):
        for item in self.r.keys("devr:2fa:%s:*" % user.username):
            self.r.delete(item)


try:
    authentication_state = AuthenticationState()
except Exception:
    log.exception("Could not initialize Duo 2fa plugin")

command_re = re.compile(r"/api/v(\d+)/cblr/session/(\d+)/command")
sensor_re = re.compile(r"/api/v(\d+)/sensor/(\d+)")


def api_route_filter(path, method, request_data):
    if is_liveresponse_api_route(path, method, request_data):
        return True
    if is_auth_api_route(path, method):
        return True

    return False


def is_auth_api_route(path, method):
    if path.startswith('/api/auth'):
        return True

    return False


def is_liveresponse_api_route(path, method, request_data):
    if command_re.match(path) and method == 'POST':
        return True
    if sensor_re.match(path) and method == 'PUT' and 'network_isolation_enabled' in request_data:
        return True

    return False


def is_logout_route(path, method):
    if method.startswith("/api/logout"):
        return True


def do_authenticate(user_id, user_name, path, method, api_token, requesting_ip, request_data):
    with DbSessionContext(cb_config) as db_session:
        user = db_session.query(User).filter(User.id == int(user_id)).filter(
            User.deleted == False).first()  # noqa: E711, E712
        if not user:
            raise AuthenticationError("Could not find user with uid of %d" % user_id)

        if api_route_filter(path, method, request_data):
            return authentication_state.perform_2fa(user, path, method,
                                                    requesting_ip.split(',')[0], request_data)
        elif is_logout_route(path, method):
            authentication_state.clear_2fa(user)

    return True


def authenticate(user_id, user_name, path, method, api_token, requesting_ip, request_data):
    try:
        retval = do_authenticate(user_id, user_name, path, method, api_token, requesting_ip.replace("::ffff:", ""),
                                 request_data)
    except AuthenticationError as e_:
        log.error(str(e_))
        return False
    except Exception:
        log.exception("Exception when processing authentication request for %s to %s %s" % (user_name,
                                                                                            method,
                                                                                            path))
        return False
    else:
        if retval is True:
            return True

    log.error("Unknown authentication error for %s to %s %s" % (user_name, method, path))
    return False
